// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <fstream>
#include "grin.h"
#include "ctree.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace std;

class tty_grin_addresses_t : public ctree {
public:
  tty_grin_addresses_t(void);
  virtual ~tty_grin_addresses_t(void);

protected:
  virtual void draw(int, int, int, int);
  virtual bool keypress(int);
  virtual void winch(void);

private:
  void resize(void);
  void commit(const ctree_node *);

  bool command_select(const char *);
  bool command_modify(const char *);
  bool command_new_address(const char *);
  bool command_new_book(const char *);
  bool command_delete(const char *);
  bool command_rename(const char *);
  bool command_help(const char *);
};

void
tty_grin_addresses_t::commit(const ctree_node *n) {
  if(n->get_data()) {
    string name=getenv("HOME");
    name+="/.ttygrin/addresses:";
    name+=n->get_text(0);
    ofstream out(name.c_str());
    if(out.good()) {
      for(const ctree_node *a=n->get_first();a;a=a->get_next()) {
	out<<a->get_text(0)<<'\t';
	out<<a->get_text(1)<<'\t';
	out<<a->get_text(2)<<endl;
      }
      out.close();
    } else report_error(strerror(errno));
  }
}

tty_grin_addresses_t::~tty_grin_addresses_t(void) {
  for(const ctree_node *n=get_first();n;n=n->get_next()) commit(n);
}

tty_grin_addresses_t::tty_grin_addresses_t(void) : ctree(3) {
  int f=open(".", O_RDONLY, 0);

  if(!(chdir(getenv("HOME"))||chdir(".ttygrin"))) {
    DIR *d=opendir(".");
    if(d) {
      struct dirent *ent;
      while((ent=readdir(d))) {
	if(!strncmp(ent->d_name, "addresses:", 10)) {
	  if(!access(ent->d_name, R_OK)) {
	    ifstream in(ent->d_name);
	    ctree_node *book=new ctree_node(*this);
	    book->set_text(0, ent->d_name+10);
	    
	    char buf[1024];
	    
	    in.getline(buf, sizeof(buf));
	    while(!(in.eof()||in.fail())) {
	      ctree_node *a=new ctree_node(*this, 0, book);
	      char *x=buf;
	      
	      for(int i=0;i<3;i++) {
		char *y=strsep(&x, "\t");
		if(y) a->set_text(i, y);
	      }
	      in.getline(buf, sizeof(buf));
	    }
	    in.close();
	  }
	}
      }
      closedir(d);
    }
  }

  fchdir(f);
  close(f);

  resize();
}

void
tty_grin_addresses_t::winch(void) {
  SLtt_get_screen_size();
  resize();
  ctree::winch();
}

void
tty_grin_addresses_t::resize(void) {
  set_width(0, SLtt_Screen_Cols/3);
  set_width(1, SLtt_Screen_Cols/3);
  set_width(2, SLtt_Screen_Cols/3);
}

void
tty_grin_addresses_t::draw(int r, int c, int rows, int cols) {
  SLsmg_set_color(SLC_HEADERS);
  SLsmg_fill_region(r, c, 1, cols, ' ');
  SLsmg_gotorc(r, 6);
  SLsmg_write_string(_("Key"));
  SLsmg_gotorc(r, 6+get_width(0));
  SLsmg_write_string(_("Comment"));
  SLsmg_gotorc(r, 6+get_width(0)+get_width(1));
  SLsmg_write_string(_("Address"));

  SLsmg_set_color(SLC_STATUS);
  SLsmg_fill_region(r+rows-2, c, 1, cols, '-');
  SLsmg_gotorc(r+rows-2, c+4);
  SLsmg_printf(_(" Address books "));

  SLsmg_set_color(SLC_MAIN);
  SLsmg_fill_region(r+rows-1, c, 1, cols, ' ');

  ctree::draw(r+1, c, rows-3, cols);
}

bool
tty_grin_addresses_t::keypress(int key) {
  const char *cmd=kbd_translate("addresses", key);

  static struct {
    const char *command;
    bool (tty_grin_addresses_t::*func)(const char *);
  } commands[]={
    {"select",         &tty_grin_addresses_t::command_select},
    {"modify-key",     &tty_grin_addresses_t::command_modify},
    {"modify-comment", &tty_grin_addresses_t::command_modify},
    {"modify-address", &tty_grin_addresses_t::command_modify},
    {"new-address",    &tty_grin_addresses_t::command_new_address},
    {"new-book",       &tty_grin_addresses_t::command_new_book},
    {"delete",         &tty_grin_addresses_t::command_delete},
    {"rename",         &tty_grin_addresses_t::command_rename},
    {"help",           &tty_grin_addresses_t::command_help},
    {0, 0}
  };

  if(cmd) {
    int i=0;
    while(commands[i].command&&
	  strcasecmp(commands[i].command, cmd)) i++;
    
    if(commands[i].command) return (this->*commands[i].func)(cmd);
  }

  return ctree::keypress(key);
}

bool
tty_grin_addresses_t::command_help(const char *cmd) {
  static tty_grin_help_key_t help_keys[]={
    {"addresses", "modify-key",     N_("Modify key")},
    {"addresses", "modify-comment", N_("Modify comment")},
    {"addresses", "modify-address", N_("Modify address")},
    {"addresses", "new-address",    N_("New entry")},
    {"addresses", "new-book",       N_("New book")},
    {"addresses", "delete",         N_("Delete")},
    {"addresses", "rename",         N_("Rename book")},
    {"tree",      "collapse",       N_("Collapse")},
    {"tree",      "expand",         N_("Expand")},
    {"tree",      "collapse-all",   N_("Collapse all")},
    {"tree",      "expand-all",     N_("Expand all")},
    {"default",   "quit",           N_("Back")},
    {0, 0, 0}
  };

  static tty_grin_help_t help={
    N_("Address books"),
    N_("Manage lists (books) of addresses to which mail can be sent.  The "
       "comment and address fields will be included in the email, the key "
       "field is used only as an identifier."),
    help_keys
  };

  tty_grin_help(help);
  return true;
}

bool
tty_grin_addresses_t::command_rename(const char *cmd) {
  if(get_cursor()&&!get_cursor()->get_parent()) {
    char buf[128];
    if(read_line(-1, 0, -1, "Rename to: ", buf, sizeof(buf))) {
      string oldname(getenv("HOME"));
      oldname+="/.ttygrin/addresses:";
      string newname=oldname;
      oldname+=get_cursor()->get_text(0);
      newname+=buf;
      if(rename(oldname.c_str(), newname.c_str()))
	report_error(strerror(errno));
      else get_cursor()->set_text(0, buf);
    }
  }
  return true;
}

bool
tty_grin_addresses_t::command_delete(const char *cmd) {
  if(get_cursor()) {
    if(get_cursor()->get_parent()) {
      if(yes_or_no(_("Delete this address (yes or no)? "))) {
	get_cursor()->get_parent()->set_data(const_cast<char *>("saveme"));
	delete get_cursor();
      }
    } else if(yes_or_no(_("Delete this address book (yes or no)? "))) {
      string name(getenv("HOME"));
      name+="/.ttygrin/addresses:";
      name+=get_cursor()->get_text(0);
      unlink(name.c_str());
      delete get_cursor();
    }
  }
  return true;
}

bool
tty_grin_addresses_t::command_new_book(const char *cmd) {
  char buf[128];

  if(read_line(-1, 0, -1, _("Book name: "), buf, sizeof(buf))) {
    string name(getenv("HOME"));
    name+="/.ttygrin/addresses:";
    name+=buf;
    
    int i=open(name.c_str(), O_WRONLY|O_CREAT, 0600);
    if(i<0) report_error(strerror(errno));
    else {
      close(i);
      ctree_node *b=new ctree_node(*this);
      b->set_text(0, buf);
    }
  }
  return true;
}

bool
tty_grin_addresses_t::command_select(const char *cmd) {
  if(get_cursor()) {
    message msg;
    string to="$";
    
    prepare_message(msg.root(), msg.root());
    if(get_cursor()->get_parent()) {
      commit(get_cursor()->get_parent());
      to+=get_cursor()->get_parent()->get_text(0);
      to+=':';
      to+=get_cursor()->get_text(0);
    } else {
      commit(get_cursor());
      to+=get_cursor()->get_text(0);
    }
    msg.root().headers()["To"]=to;
    tty_grin_compose(msg);
    return false;
  }
  return true;
}

bool
tty_grin_addresses_t::command_modify(const char *cmd) {
  if(get_cursor()&&get_cursor()->get_parent()) {
    int which=0;
    switch(tolower(cmd[7])) {
    case 'k': which=0; break;
    case 'c': which=1; break;
    case 'a': which=2; break;
    }
    static char *prompt[]={N_("Key: "), N_("Comment: "), N_("Address: ")};
    char buf[128];
    if(read_line(-1, 0, -1, gettext(prompt[which]), buf, sizeof(buf))) {
      get_cursor()->set_text(which, buf);
      get_cursor()->get_parent()->set_data(const_cast<char *>("saveme"));
    }
  }
  return true;
}

bool
tty_grin_addresses_t::command_new_address(const char *cmd) {
  if(get_cursor()) {
    ctree_node *p=get_cursor();
    if(p->get_parent()) p=p->get_parent();
    
    char key[128];
    char comment[128];
    char address[128];
    if(read_line(-1, 0, -1, _("Key: "), key, sizeof(key))&&
       read_line(-1, 0, -1, _("Comment: "), comment, sizeof(comment))&&
       read_line(-1, 0, -1, _("Address: "), address, sizeof(address))) {
      ctree_node *n=new ctree_node(*this, 0, p);
      n->set_text(0, key);
      n->set_text(1, comment);
      n->set_text(2, address);
      p->set_data(const_cast<char *>("saveme"));
    }
  }
  return true;
}

class tty_grin_pickaddr_t : public tty_grin_addresses_t {
public:
  tty_grin_pickaddr_t(void) { result=0; }
  virtual ~tty_grin_pickaddr_t(void);
  const ctree_node *get_result(void) const { return result; }

protected:
  bool keypress(int);

private:
  ctree_node *result;
};

tty_grin_pickaddr_t::~tty_grin_pickaddr_t(void) {
}

bool
tty_grin_pickaddr_t::keypress(int key) {
  const char *cmd=kbd_translate("addresses", key);

  if(cmd&&!strcasecmp(cmd, "select")&&get_cursor()) {
    result=get_cursor();
    return false;
  }

  return ctree::keypress(key);
}

void
tty_grin_addresses(void) {
  tty_grin_addresses_t n;

  n.set_cursor_color(SLC_CURSOR);
  n.run();
}

bool
tty_grin_pickaddr(char *buf, size_t maxsize) {
  tty_grin_pickaddr_t n;

  n.set_cursor_color(SLC_CURSOR);
  n.run();

  if(n.get_result()) {
    if(n.get_result()->get_parent())
      snprintf(buf, maxsize, "$%s:%s",
	       n.get_result()->get_parent()->get_text(0),
	       n.get_result()->get_text(0));
    else snprintf(buf, maxsize, "$%s", n.get_result()->get_text(0));
    return true;
  }

  return false;
}
