// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <errno.h>
#include "localmailbox.h"
#include "grin.h"
#include "ctree.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace std;

class tty_grin_boxlist_t : public ctree {
public:
  tty_grin_boxlist_t(boxlist&, const char *);

protected:
  void draw(int, int, int, int);
  bool keypress(int);
  void winch(void);

private:
  boxlist& bl;
  const char *nm;

  void resize(void);

  bool command_select(const char *);
  bool command_create(const char *);
  bool command_delete(const char *);
  bool command_rename(const char *);
  bool command_help(const char *);
};

tty_grin_boxlist_t::tty_grin_boxlist_t(boxlist& b, const char *n) :
  ctree(1), bl(b) {
  nm=n;

  for(int i=0;i<b.size();i++) {
    ctree_node *n=new ctree_node(*this);
    n->set_text(0, b[i]);
  }

  resize();
}

void
tty_grin_boxlist_t::winch(void) {
  SLtt_get_screen_size();
  resize();
  ctree::winch();
}

void
tty_grin_boxlist_t::resize(void) {
  set_width(0, SLtt_Screen_Cols);
}

void
tty_grin_boxlist_t::draw(int r, int c, int rows, int cols) {
  SLsmg_set_color(SLC_STATUS);
  SLsmg_fill_region(r+rows-2, c, 1, cols, '-');
  SLsmg_gotorc(r+rows-2, c+4);
  SLsmg_printf(_(" %s -- %d mailboxes "), nm, bl.size());

  SLsmg_set_color(SLC_MAIN);
  SLsmg_fill_region(r+rows-1, c, 1, cols, ' ');

  ctree::draw(r, c, rows-2, cols);
}

bool
tty_grin_boxlist_t::keypress(int key) {
  const char *cmd=kbd_translate("boxlist", key);

  static struct {
    const char *command;
    bool (tty_grin_boxlist_t::*func)(const char *);
  } commands[]={
    {"select", &tty_grin_boxlist_t::command_select},
    {"create", &tty_grin_boxlist_t::command_create},
    {"delete", &tty_grin_boxlist_t::command_delete},
    {"rename", &tty_grin_boxlist_t::command_rename},
    {"help",   &tty_grin_boxlist_t::command_help},
    {0, 0}
  };

  if(cmd) {
    int i=0;
    while(commands[i].command&&
          strcasecmp(commands[i].command, cmd)) i++;
    
    if(commands[i].command) return (this->*commands[i].func)(cmd);
  }

  return ctree::keypress(key);
}

void
tty_grin_boxlist(boxlist& b, const char *n) {
  tty_grin_boxlist_t bl(b, n);

  bl.set_cursor_color(SLC_CURSOR);
  bl.run();
}

bool
tty_grin_boxlist_t::command_create(const char *cmd) {
  char buf[128];

  if(read_line(-1, 0, -1, _("Mailbox name: "), buf, sizeof(buf))) {
    if(bl.create(buf)) (new ctree_node(*this))->set_text(0, buf);
    else report_error(strerror(errno));
  }
  return true;
}

bool
tty_grin_boxlist_t::command_delete(const char *cmd) {
  if(get_cursor()&&
     yes_or_no(_("Remove this mailbox (yes or no)? "))) {
    if(bl.remove(get_cursor()->get_text(0))) delete get_cursor();
      else report_error(strerror(errno));
  }
  return true;
}

bool
tty_grin_boxlist_t::command_rename(const char *cmd) {
  char buf[128];

  if(get_cursor()&&
     read_line(-1, 0, -1, _("New name: "), buf, sizeof(buf))) {
    if(bl.move(get_cursor()->get_text(0), buf))
      get_cursor()->set_text(0, buf);
    else report_error(strerror(errno));
  }
  return true;
}

bool
tty_grin_boxlist_t::command_select(const char *cmd) {
  if(get_cursor()) {
    mailbox *m=bl.openbox(get_cursor()->get_text(0));
    if(m) {
      localmailbox *b=dynamic_cast<localmailbox *>(m);
      if(!b||b->lock())
	tty_grin_mailbox(*m, get_cursor()->get_text(0));
      else report_error(_("Mailbox is currently in use"));
      delete m;
    } else report_error(strerror(errno));
  }
  return true;
}

bool
tty_grin_boxlist_t::command_help(const char *cmd) {
  static tty_grin_help_key_t help_keys[]={
    {"boxlist", "create", N_("Create")},
    {"boxlist", "delete", N_("Delete")},
    {"boxlist", "rename", N_("Rename")},
    {"default", "quit",   N_("Back")},
    {0, 0, 0}
  };

  static tty_grin_help_t help={
    N_("Mailboxes"),
    N_("Create, delete, rename and examine mailboxes, either local or "
       "remote."),
    help_keys
  };

  tty_grin_help(help);
  return true;
}
