// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <signal.h>
#include <regex.h>
#include "grin.h"
#include "ctree.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace std;

ctree_node::ctree_node(ctree& o, ctree_node *n,
ctree_node *p) {
  expanded=false;
  first=last=0;
  text=new string[o.columns];
  owner=&o;
  parent=p;
  next=n;

  if(next) prev=next->prev;
  else if(parent) prev=parent->last;
  else prev=owner->last;

  if(next) next->prev=this;
  if(prev) prev->next=this;

  if(parent) {
    if(parent->first==next) parent->first=this;
    if(parent->last==prev) parent->last=this;
  } else {
    if(owner->first==next) owner->first=this;
    if(owner->last==prev) owner->last=this;
  }
}

ctree_node::~ctree_node(void) {
  delete[] text;
  if(next) next->prev=prev;
  if(prev) prev->next=next;
  if(parent) {
    if(parent->first==this) parent->first=next;
    if(parent->last==this) parent->last=prev;
  } else {
    if(owner->first==this) owner->first=next;
    if(owner->last==this) owner->last=prev;
  }
  if(owner->cursor==this) owner->cursor=first;
  if(owner->scroll==this) owner->scroll=first;
  while(first) delete first;
}

void
ctree_node::move(ctree_node *n, ctree_node *p) {
  for(ctree_node *x=n;x;x=x->get_parent())
    if(x==this) return;
  
  if(next) next->prev=prev;
  if(prev) prev->next=next;

  if(parent) {
    if(parent->first==this) parent->first=next;
    if(parent->last==this) parent->last=prev;
  } else {
    if(owner->first==this) owner->first=next;
    if(owner->last==this) owner->last=prev;
  }

  if(owner->cursor==this) owner->cursor=first;
  if(owner->scroll==this) owner->scroll=first;

  parent=p;
  next=n;

  if(next) prev=next->prev;
  else if(parent) prev=parent->last;
  else prev=owner->last;

  if(next) next->prev=this;
  if(prev) prev->next=this;

  if(parent) {
    if(parent->first==next) parent->first=this;
    if(parent->last==prev) parent->last=this;
  } else {
    if(owner->first==next) owner->first=this;
    if(owner->last==prev) owner->last=this;
  }
}

const ctree_node *
ctree_node::up(bool e) const {
  if(prev) {
    const ctree_node *u=prev;
    while((e||u->expanded)&&u->last) u=u->last;
    return u;
  }
  return parent;
}

const ctree_node *
ctree_node::down(bool e) const {
  if((e||expanded)&&first) return first;
  if(next) return next;

  const ctree_node *d=parent;
  while(d) {
    if(d->next) return d->next;
    d=d->parent;
  }
  return 0;
}

ctree::ctree(unsigned int w) {
  columns=w;
  widths=new unsigned int[w];
  first=last=cursor=scroll=0;
  textc=cursorc=0;
  cursor_char='>';
  lastsearch[0]=0;
}

ctree::~ctree(void) {
  delete[] widths;
  while(first) delete first;
}

void
ctree::run(void) {
  void *sigdata_old=sigdata_winch;
  sigdata_winch=this;
  SLSig_Fun_Type *f=SLsignal(SIGWINCH, sigwinch);

  do {
    draw(0, 0, SLtt_Screen_Rows, SLtt_Screen_Cols);
    SLsmg_refresh();
  } while(keypress(SLkp_getkey()));

  sigdata_winch=sigdata_old;
  SLsignal(SIGWINCH, f);
}

void
ctree::draw(int r, int c, int rows, int cols) {
  SLsmg_set_color(textc);
  SLsmg_fill_region(r, c, rows, cols, ' ');

  if(!first) return;

  {
    if(!cursor) cursor=first;
    ctree_node *n;

    n=cursor;
    for(int i=1;i<rows&&n&&n!=scroll;i++) n=n->up();
    if(!n||n!=scroll) scroll=cursor;
  }

  const ctree_node *n=scroll;
  int cline=-1;
  for(int i=0;i<rows&&n;i++) {
    if(n==cursor) cline=i;
    int total=4;
    for(const ctree_node *x=n;x;x=x->get_parent()) total+=2;
    SLsmg_gotorc(r+i, c+total-4);
    SLsmg_set_color(cursorc);
    SLsmg_write_string
      (const_cast<char *>(n->get_first()?
			  (n->get_expanded()?"[-] ":"[+] "):"    "));
    SLsmg_set_color(textc);
    for(int j=0;j<(signed)columns&&total<cols;j++) {
      int len=get_width(j);
      if(len>cols-total) len=cols-total;
      SLsmg_gotorc(r+i, c+total);
      SLsmg_write_nstring((char *)n->get_text(j), len-1);
      total+=len;
    }
    n=n->down();
  }

  SLsmg_gotorc(r+cline, 0);
  if(cline>=0) {
    SLsmg_set_color(cursorc);
    SLsmg_write_char(cursor_char);
    SLsmg_set_color(textc);
  }
}

void
ctree::winch(void) {
  SLtt_get_screen_size();
  SLsmg_reinit_smg();
  draw(0, 0, SLtt_Screen_Rows, SLtt_Screen_Cols);
  SLsmg_refresh();
}

void
ctree::set_all(ctree_node *n, bool expanded) {
  if(n->get_first()) {
    n->set_expanded(expanded);
    for(ctree_node *x=n->get_first();x;x=x->get_next()) set_all(x, expanded);
  }
}

bool
ctree::keypress(int key) {
  const char *cmd=kbd_translate("tree", key);

  static struct {
    const char *command;
    bool (ctree::*func)(const char *);
  } commands[]={
    {"collapse",     &ctree::command_collapse},
    {"collapse-all", &ctree::command_all},
    {"down",         &ctree::command_down},
    {"expand",       &ctree::command_expand},
    {"expand-all",   &ctree::command_all},
    {"find-back",    &ctree::command_find},
    {"find-forward", &ctree::command_find},
    {"jump-end",     &ctree::command_jump_end},
    {"jump-top",     &ctree::command_jump_top},
    {"page-down",    &ctree::command_page_down},
    {"page-up",      &ctree::command_page_up},
    {"quit",         &ctree::command_quit},
    {"redraw",       &ctree::command_redraw},
    {"up",           &ctree::command_up},
    {0, 0}
  };

  if(cmd) {
    int i=0;
    while(commands[i].command&&
	  strcasecmp(commands[i].command, cmd)) i++;
    
    if(commands[i].command) return (this->*commands[i].func)(cmd);
  }

  SLtt_beep();
  return true;
}

void
ctree::sigwinch(int sig) {
  ((ctree *)sigdata_winch)->winch();
}

bool
ctree::command_find(const char *cmd) {
  char next[1024];
  if(read_line(-1, 0, -1, _("Find: "), next, sizeof(next))) {
    regex_t preg;
    
    if(regcomp(&preg, next[0]?next:lastsearch,
	       REG_EXTENDED|REG_ICASE|REG_NOSUB)) SLtt_beep();
    else {
      ctree_node *n=cursor;
      unsigned int i=columns;
      
      while(n&&i>=columns) {
	n=tolower(cmd[5])=='f'?n->down(true):n->up(true);
	if(n) {
	  i=0;
	  while(i<columns&&regexec(&preg, n->get_text(i), 0, 0, 0)) i++;
	}
      }
      
      if(n) {
	cursor=n;
	for(ctree_node *n=get_cursor()->get_parent();n;n=n->get_parent())
	  n->set_expanded(true);
      } else SLtt_beep();
    }
    
    if(next[0]) strcpy(lastsearch, next);
  }

  return true;
}

bool
ctree::command_all(const char *cmd) {
  for(ctree_node *n=first;n;n=n->get_next()) set_all(n, tolower(*cmd)=='e');
  if(tolower(*cmd)=='c'&&get_cursor()) {
    for(ctree_node *n=get_cursor()->get_parent();n;n=n->get_parent())
      n->set_expanded(true);
  }

  return true;
}

bool
ctree::command_up(const char *cmd) {
  if(cursor&&cursor->up()) cursor=cursor->up();
  else SLtt_beep();
  return true;
}

bool
ctree::command_down(const char *cmd) {
  if(cursor&&cursor->down()) cursor=cursor->down();
  else SLtt_beep();
  return true;
}

bool
ctree::command_page_up(const char *cmd) {
  if(cursor&&cursor->up())
    for(int i=0;i<SLtt_Screen_Rows&&cursor->up();i++)
      cursor=cursor->up();
  else SLtt_beep();
  return true;
}

bool
ctree::command_page_down(const char *cmd) {
  if(cursor&&cursor->down())
    for(int i=0;i<SLtt_Screen_Rows&&cursor->down();i++)
      cursor=cursor->down();
  else SLtt_beep();
  return true;
}

bool
ctree::command_jump_top(const char *cmd) {
  cursor=first;
  return true;
}

bool
ctree::command_jump_end(const char *cmd) {
  cursor=last;
  if(cursor) while(cursor->get_expanded()&&cursor->get_last())
    cursor=cursor->get_last();
  return true;
}

bool
ctree::command_collapse(const char *cmd) {
  if(cursor) cursor->set_expanded(false);
  else SLtt_beep();
  return true;
}

bool
ctree::command_expand(const char *cmd) {
  if(cursor&&cursor->get_first())
    cursor->set_expanded(true);
  else SLtt_beep();
  return true;
}

bool
ctree::command_redraw(const char *cmd) {
  SLsmg_touch_lines(0, SLtt_Screen_Rows);
  return true;
}

bool
ctree::command_quit(const char *cmd) {
  return false;
}
