// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#include <fstream>
#include "grin.h"
#include "ctree.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace std;

class tty_grin_groupbrowse_t : public ctree {
public:
  tty_grin_groupbrowse_t(const char *s) : ctree(1) {
    server=s;
    result=0;
    set_width(0, SLtt_Screen_Cols);
  }
  const char *get_result(void) const { return result; }

protected:
  void winch(void);
  void draw(int, int, int, int);
  bool keypress(int);

private:
  const char *server, *result;
};

void
tty_grin_groupbrowse_t::winch(void) {
  SLtt_get_screen_size();
  set_width(0, SLtt_Screen_Cols);
  ctree::winch();
}

void
tty_grin_groupbrowse_t::draw(int r, int c, int rows, int columns) {
  SLsmg_set_color(SLC_STATUS);
  SLsmg_fill_region(r+rows-2, c, 1, columns, '-');
  SLsmg_gotorc(r+rows-2, c+4);
  SLsmg_printf(_(" Newsgroups on %s "), server);

  SLsmg_set_color(SLC_MAIN);
  SLsmg_fill_region(r+rows-1, c, 1, columns, ' ');

  ctree::draw(r, c, rows-2, columns);
}

bool
tty_grin_groupbrowse_t::keypress(int key) {
  tty_grin_help_key_t help_keys[]={
    {"tree",    "collapse",     N_("Collapse")},
    {"tree",    "expand",       N_("Expand")},
    {"tree",    "collapse-all", N_("Collapse all")},
    {"tree",    "expand-all",   N_("Expand all")},
    {"default", "quit",         N_("Back")},
    {0, 0, 0}
  };

  tty_grin_help_t help={
    N_("Newsgroup browser"),
    N_("Select available news groups interactively from the news server's "
       "list."),
    help_keys
  };

  const char *cmd=kbd_translate("groupbrowse", key);

  if(cmd) {
    if(!strcasecmp(cmd, "select")) {
      if(get_cursor()&&!get_cursor()->get_first()) {
	result=get_cursor()->get_text(0);
	return false;
      }
    } else if(!strcasecmp(cmd, "help"))
      tty_grin_help(help);
    else return ctree::keypress(key);
  } else return ctree::keypress(key);

  return true;
}

bool
tty_grin_groupbrowse(const char *server, char *outbuf, size_t maxsize) {
  char buf[1024], *port;
  strncpy(buf, server, sizeof(buf));
  buf[sizeof(buf)-1]=0;
  if((port=strchr(buf, ':'))) *(port++)=0;
  else port="119";

  struct hostent *ent;
  if(!(ent=gethostbyname(buf))) {
    report_error(hstrerror(h_errno));
    return false;
  }

  int sock;
  if((sock=socket(PF_INET, SOCK_STREAM, 0))<0) {
    report_error(strerror(errno));
    return false;
  }

  struct sockaddr_in addr;
  addr.sin_family=AF_INET;
  addr.sin_port=htons(atoi(port));
  addr.sin_addr=*(struct in_addr *)ent->h_addr;

  if(connect(sock, (struct sockaddr *)&addr, sizeof(addr))<0) {
    report_error(strerror(errno));
    close(sock);
    return false;
  }

  fstream nntp(sock);

  nntp.getline(buf, sizeof(buf));
  if(atoi(buf)!=200) {
    report_error(buf);
    nntp.close();
    return false;
  }

  nntp<<"LIST\r\n";
  nntp.flush();

  nntp.getline(buf, sizeof(buf));
  if(atoi(buf)!=215) {
    report_error(buf);
    nntp.close();
    return false;
  }

  tty_grin_groupbrowse_t browse(server);
  browse.set_cursor_color(SLC_CURSOR);

  nntp.getline(buf, sizeof(buf));
  while(!(nntp.eof()||nntp.fail())&&buf[0]!='.') {
    char *end=strchr(buf, ' ');

    if(end) *end=0;

    int j=0;
    ctree_node *n=0;

    while(buf[j]!='.'&&buf[j]) j++;
    while(buf[j]) {
      j++;
      string sub(buf, 0, j);

      ctree_node *parent=n?n->get_last():browse.get_last();
      if(parent&&parent->get_text(0)==sub) n=parent;
      else (n=new ctree_node(browse, 0, n))->set_text(0, sub.c_str());

      while(buf[j]!='.'&&buf[j]) j++;
    }

    (n=new ctree_node(browse, 0, n))->set_text(0, buf);

    nntp.getline(buf, sizeof(buf));    
  }

  nntp<<"QUIT\r\n";
  nntp.flush();
  nntp.close();

  browse.run();

  if(browse.get_result()) {
    strncpy(outbuf, browse.get_result(), maxsize);
    outbuf[maxsize-1]=0;
    return true;
  }

  return false;
}
