// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <string.h>
#include <ctype.h>
#include <map>
#include <string>
#include "grin.h"
#include "keybind.h"

using namespace std;

struct {
  int code;
  const char *name;
} kbd_special_names[]={
  {9,                "Tab"},
  {127,              "^?"},
  {32,               "Space"},
  {SL_KEY_UP,        "Up"},
  {SL_KEY_DOWN,      "Down"},
  {SL_KEY_LEFT,      "Left"},
  {SL_KEY_RIGHT,     "Right"},
  {SL_KEY_PPAGE,     "PgUp"},
  {SL_KEY_NPAGE,     "PgDown"},
  {SL_KEY_HOME,      "Home"},
  {SL_KEY_END,       "End"},
  {SL_KEY_A1,        "A1"},
  {SL_KEY_A3,        "A3"},
  {SL_KEY_B2,        "B2"},
  {SL_KEY_C1,        "C1"},
  {SL_KEY_C3,        "C3"},
  {SL_KEY_REDO,      "Redo"},
  {SL_KEY_UNDO,      "Undo"},
  {SL_KEY_BACKSPACE, "Backspace"},
  {SL_KEY_ENTER,     "Enter"},
  {SL_KEY_IC,        "IC"},
  {SL_KEY_DELETE,    "Del"},
  {0, 0}
};

struct kbd_domain {
  map<int, string> lookup;
  map<string, int> reverse;
};

map<string, kbd_domain> kbd;

int
kbd_name_to_key(const char *n) {
  if(!n[0]) return -1;
  if(!n[1]) return n[0];

  if(n[0]=='^'&&isalpha(n[1])&&!n[2]) return tolower(n[1])-'a'+1;
  if(tolower(n[0])=='f'&&isdigit(n[1])) return atoi(n+1)+SL_KEY_F0;

  for(int i=0;kbd_special_names[i].name;i++)
    if(!strcasecmp(n, kbd_special_names[i].name))
      return kbd_special_names[i].code;

  return SL_KEY_ERR;
}

string
kbd_key_to_name(int key) {
  for(int i=0;kbd_special_names[i].name;i++)
    if(key==kbd_special_names[i].code)
      return kbd_special_names[i].name;

  if(key<32) {
    string n="^";
    n+=(char)(key-1+'A');
    return n;
  }

  if(key>=SL_KEY_F0) {
    char buf[16];

    snprintf(buf, sizeof(buf), "F%d", key-SL_KEY_F0);
    return buf;
  }

  char x[]={key, 0};

  return x;
}

void
kbd_bind(const char *domain, int key, const char *command) {
  if(key>='a'&&key<='z') key-=32;
  kbd[domain].lookup[key]=command;
  kbd[domain].reverse[command]=key;
}

const char *
kbd_translate_wrap(const char *domain, int key) {
  map<string, kbd_domain>::const_iterator
    fd=kbd.find(domain);

  if(fd==kbd.end()) return 0;

  map<int, string>::const_iterator
    i=fd->second.lookup.find(key);

  if(i==fd->second.lookup.end()) return 0;

  return i->second.c_str();
}

const char *
kbd_translate(const char *domain, int key) {
  if(key>='a'&&key<='z') key-=32;
  const char *x=kbd_translate_wrap(domain, key);

  if(x) return x;

  return kbd_translate_wrap("default", key);
}

string
kbd_reverse_translate(const char *domain, const char *command) {
  map<string, kbd_domain>::const_iterator fd=kbd.find(domain);

  if(fd==kbd.end()) return 0;

  map<string, int>::const_iterator
    i=fd->second.reverse.find(command);

  if(i==fd->second.reverse.end()) return "";

  return kbd_key_to_name(i->second);
}

void
kbd_init(void) {
  for(int i=0;kbd_default[i].name;i++)
    for(int j=0;kbd_default[i].bindings[j].key;j++)
      kbd_bind(kbd_default[i].name, kbd_default[i].bindings[j].key,
	       kbd_default[i].bindings[j].command);
}

void
kbd_dump(ostream& out) {
  for(map<string, kbd_domain>::const_iterator i=kbd.begin();
      i!=kbd.end();++i) {
    for(map<int, string>::const_iterator j=i->second.lookup.begin();
	j!=i->second.lookup.end();++j)
      out<<"bind "<<i->first<<' '<<
	kbd_key_to_name(j->first)<<' '<<j->second<<endl;
    out<<endl;
  }
}
