// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#define _GNU_SOURCE
#include <fstream>
#include <time.h>
#include "grin.h"
#include "ctree.h"
#include "tempfile.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace std;

class tty_grin_mailbox_t : public ctree {
public:
  tty_grin_mailbox_t(mailbox&, const char *);

protected:
  virtual void winch(void);
  virtual bool keypress(int);
  virtual void draw(int, int, int, int);

private:
  void resize(void);
  void flagify(ctree_node *);
  char flagmap(int);

  const char *name;
  mailbox *ng;

  bool command_toggle(const char *cmd);
  bool command_select(const char *cmd);
  bool command_help(const char *cmd);
  bool command_quit(const char *cmd);
};

bool
tty_grin_mailbox_t::command_select(const char *cmd) {
  if(get_cursor()) {
    message_descriptor *m=(message_descriptor *)get_cursor()->get_data();
    tempfile t;
    
    ofstream out(t.filename());
    if(out.good()) {
      m->extract(out);
      out.close();
      ifstream in(t.filename());
      if(in.good()) {
	message msg(in);
	in.close();
	m->set_flag(message_descriptor::READ);
	tty_grin_message(msg, m);
	flagify(get_cursor());
      }
    }      
  }

  return true;
}

bool
tty_grin_mailbox_t::command_quit(const char *cmd) {
  ctree_node *n;
  for(n=get_first();n&&
	!(((message_descriptor *)n->get_data())->get_flags()&
	  message_descriptor::DELETED);
      n=n->get_next());
  if(n&&yes_or_no(_("Expunge deleted messages (yes or no)? ")))
    ng->expunge();

  return false;
}

bool
tty_grin_mailbox_t::command_help(const char *cmd) {
  static tty_grin_help_key_t help_keys[]={
    {"mailbox", "toggle-read",    N_("Toggle read")},
    {"mailbox", "toggle-marked",  N_("Toggle special")},
    {"mailbox", "toggle-deleted", N_("Toggle deleted")},
    {"default", "find-forward",   N_("Search forward")},
    {"default", "find-back",      N_("Search back")},
    {"default", "quit",           N_("Back")},
    {0, 0, 0}
  };

  static tty_grin_help_t help={
    N_("Mailbox"),
    N_("Read email messages stored in local or remote mailboxes.  Message "
       "markings appear as follows:\n\n"
       "  R - New/unread\n"
       "  A - Answered\n"
       "  * - Special\n"
       "  D - Deleted"),
    help_keys
  };

  tty_grin_help(help);
  return true;
}

bool
tty_grin_mailbox_t::command_toggle(const char *cmd) {
  int flag=0;

  switch(tolower(cmd[7])) {
  case 'r': flag=message_descriptor::READ; break;
  case 'd': flag=message_descriptor::DELETED; break;
  case 'm': flag=message_descriptor::MARKED; break;
  }

  if(get_cursor())
    ((message_descriptor *)get_cursor()->get_data())->toggle_flag(flag);
  flagify(get_cursor());
  if(get_cursor()->down()) set_cursor(get_cursor()->down());

  return true;
}

bool
tty_grin_mailbox_t::keypress(int key) {
  const char *cmd=kbd_translate("mailbox", key);

  static struct {
    const char *command;
    bool (tty_grin_mailbox_t::*func)(const char *);
  } commands[]={
    {"toggle-read",     &tty_grin_mailbox_t::command_toggle},
    {"toggle-marked",   &tty_grin_mailbox_t::command_toggle},
    {"toggle-deleted",  &tty_grin_mailbox_t::command_toggle},
    {"select",          &tty_grin_mailbox_t::command_select},
    {"help",            &tty_grin_mailbox_t::command_help},
    {"quit",            &tty_grin_mailbox_t::command_quit},
    {0, 0}
  };

  if(cmd) {
    int i=0;
    while(commands[i].command&&
          strcasecmp(commands[i].command, cmd)) i++;
    
    if(commands[i].command) return (this->*commands[i].func)(cmd);
  }

  return ctree::keypress(key);
}

void
tty_grin_mailbox_t::draw(int r, int c, int rows, int columns) {
  SLsmg_set_color(SLC_STATUS);
  SLsmg_fill_region(r+rows-2, c, 1, columns, '-');
  SLsmg_gotorc(r+rows-2, c+4);
  SLsmg_printf(_(" %s -- %d messages "), name,
	       ng->size());

  SLsmg_set_color(SLC_MAIN);
  SLsmg_fill_region(r+rows-1, c, 1, columns, ' ');

  ctree::draw(r, c, rows-2, columns);
}

char
tty_grin_mailbox_t::flagmap(int x) {
  if(x&message_descriptor::DELETED) return 'D';
  if(x&message_descriptor::MARKED) return '*';
  if(x&message_descriptor::ANSWERED) return 'A';
  if(x&message_descriptor::READ) return 0;
  return 'N';
}

void
tty_grin_mailbox_t::flagify(ctree_node *node) {
  message_descriptor *msg=(message_descriptor *)node->get_data();

  char buf[2];

  buf[0]=flagmap(msg->get_flags());
  buf[1]=0;
  node->set_text(0, buf);
}

void
tty_grin_mailbox_t::resize(void) {
  set_width(0, 5);
  set_width(1, (SLtt_Screen_Cols-25)/2);
  set_width(2, (SLtt_Screen_Cols-25)/2);
  set_width(3, SLtt_Screen_Cols);
}

void
tty_grin_mailbox_t::winch(void) {
  SLtt_get_screen_size();
  resize();
  ctree::winch();
}

tty_grin_mailbox_t::tty_grin_mailbox_t(mailbox& n, const char *nm) : ctree(4) {
  ng=&n;
  name=nm;
  
  for(int i=0;i<n.size();i++) {
    ctree_node *m=new ctree_node(*this);
    m->set_text(1, n[i].get_subject());
    m->set_text(2, n[i].get_from());
    
    const char *rawdate=n[i].get_date();
    if(strlen(rawdate)>5) {
      if(rawdate[3]==',') rawdate+=5;
      struct tm date;
      strptime(rawdate, "%d %b %Y %T", &date);
      char buf[64];
      strftime(buf, sizeof(buf), "%F", &date);
      buf[sizeof(buf)-1]=0;
      m->set_text(3, buf);
    }
    
    m->set_data((void *)&n[i]);
    flagify(m);
  }

  resize();
}

void
tty_grin_mailbox(mailbox& n, const char *nm) {
  tty_grin_mailbox_t g(n, nm);

  g.set_cursor_color(SLC_CURSOR);

  g.run();
}
