// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <list>
#include <string>
#include <fstream>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <regex.h>
#include "grin.h"
#include "ctree.h"
#include "tempfile.h"
#include "newsgroup.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace std;

struct tty_grin_line_t {
  tty_grin_line_t(int a, string x) : attr(a), text(x) { }
  tty_grin_line_t(string x) : attr(SLC_MAIN), text(x) { }
  tty_grin_line_t(const char *x) : attr(SLC_MAIN), text(x) { }

  int attr;
  string text;
};

struct tty_grin_message_t {
  list<tty_grin_line_t> lines;
  list<tty_grin_line_t>::const_iterator top;
  list<const message_part *> parts;
  unsigned int attachments;
  unsigned int linenum;
  message *msg;
  message_descriptor *desc;
  list<string> pgp_output;
  list<message_part> uuencode;
  bool rawmode;
};

class tty_grin_message_am : public ctree {
public:
  tty_grin_message_am(void) : ctree(3) { resize(); }

protected:
  virtual void winch(void);
  virtual void draw(int, int, int, int);
  virtual bool keypress(int);

private:
  void resize(void);
};

bool
tty_grin_message_am::keypress(int key) {
  static tty_grin_help_key_t help_keys[]={
    {"message", "save", N_("Save attachment")},
    {"default", "quit", N_("Back")},
    {0, 0, 0}
  };

  static tty_grin_help_t help={
    N_("Message attachments"),
    N_("View or store attachments and parts of a multi-part message"),
    help_keys
  };

  const char *cmd=kbd_translate("message", key);

  if(cmd) {
    if(!strcasecmp(cmd, "select")) {
      if(get_cursor()) {
	message_part *p=(message_part *)(get_cursor()->get_data());
	message_part::header_list::const_iterator
	  t=p->headers().find("content-type");
	const char *handler=
	  config_mime_handler(t==p->headers().end()?
			      "text/plain":t->second.c_str());
	
	if(handler) {
	  tempfile t;
	  ofstream out(t.filename());
	  
	  if(out.good()) {
	    for(unsigned int i=0;i<p->content().size();i++)
	      out<<(char)p->content()[i];
	    out.close();
	    
	    char command[1024];
	    snprintf(command, sizeof(command), "%s %s", handler, t.filename());
	    
	    SLsmg_suspend_smg();
	    system(command);
	    SLsmg_resume_smg();
	  } else report_error(strerror(errno));
	} else SLtt_beep();
      }
    } else if(!strcasecmp(cmd, "save")) {
      if(get_cursor()) {
	char buf[1024];
	if(read_line(-1, 0, -1, _("Filename: "), buf, sizeof(buf), FILENAME)&&
	   (access(buf, F_OK)||
	    yes_or_no(_("Overwrite existing file (yes or no)? ")))) {
	  ofstream out(buf);
	  
	  if(out.good()) {
	    message_part *p=(message_part *)(get_cursor()->get_data());
	    for(unsigned int i=0;i<p->content().size();i++)
	      out<<(char)p->content()[i];
	    out.close();
	  } else report_error(strerror(errno));
	}
      }
    } else if(!strcasecmp(cmd, "help")) tty_grin_help(help);
    else return ctree::keypress(key);
  }
  return ctree::keypress(key);
}

void
tty_grin_message_am::winch(void) {
  SLtt_get_screen_size();
  resize();
  ctree::winch();
}

void
tty_grin_message_am::draw(int r, int c, int rows, int cols) {
  SLsmg_set_color(SLC_STATUS);
  SLsmg_fill_region(r+rows-2, c, 1, cols, '-');
  SLsmg_gotorc(r+rows-2, c+4);
  SLsmg_write_string(_(" Attachments "));
  SLsmg_set_color(SLC_MAIN);
  SLsmg_fill_region(r+rows-1, c, 1, cols, ' ');
  ctree::draw(r, c, rows-2, cols);
}

void
tty_grin_message_am::resize(void) {
  int i=SLtt_Screen_Cols/3;

  set_width(0, i);
  set_width(1, i);
  set_width(2, i);
}

void
tty_grin_message_attachments(tty_grin_message_t& m) {
  tty_grin_message_am a;

  for(list<const message_part *>::iterator p=m.parts.begin();
      p!=m.parts.end();p++) {
    ctree_node *n=new ctree_node(a);
    message_part::header_list::const_iterator i;
    i=(*p)->headers().find("content-type.name");
    if(i==(*p)->headers().end())
      i=(*p)->headers().find("content-disposition.filename");
    if(i!=(*p)->headers().end()) n->set_text(0, i->second.c_str());
    i=(*p)->headers().find("content-type");
    n->set_text(1, i==(*p)->headers().end()?"text/plain":i->second.c_str());

    char buf[16];
    unsigned int size=(*p)->content().size();
    if(size<1024) snprintf(buf, sizeof(buf), _("%d bytes"), size);
    else snprintf(buf, sizeof(buf), _("%.01f kbytes"), ((float)size)/1024.0);
    n->set_text(2, buf);

    n->set_data((message_part *)*p);
  }

  a.set_cursor_color(SLC_CURSOR);
  a.run();
}

void
tty_grin_message_assemble(const message_part& msg, tty_grin_message_t& m) {
  if(!m.rawmode&&m.lines.empty()&&m.pgp_output.size())
    for(list<string>::const_iterator i=m.pgp_output.begin();
	i!=m.pgp_output.end();++i)
      m.lines.push_back(tty_grin_line_t(SLC_INFO, *i));

  char div[SLtt_Screen_Cols+1];
  memset(div, '-', SLtt_Screen_Cols);
  div[SLtt_Screen_Cols]=0;

  static char *show_headers[]={
    N_("Date"), N_("From"), N_("To"),
    N_("Cc"), N_("Newsgroups"), N_("Subject"),
    N_("Organization"),
    0
  };

  if(!m.rawmode) {
    bool firstheader=m.lines.size();
    for(char **x=show_headers;*x;x++) {
      message_part::header_list::const_iterator
      h=msg.headers().find(*x);
      if(h!=msg.headers().end()) {
	string header(gettext(*x));
	header+=": ";
	header+=h->second;
	if(firstheader) {
	  m.lines.push_back(tty_grin_line_t(SLC_DIVIDER, div));
	  firstheader=false;
	}
	m.lines.push_back(tty_grin_line_t(SLC_HEADERS, header));
      }
    }
  }

  message_part::header_list::const_iterator
    i=msg.headers().find("content-type");

  if(msg.children().empty()) {
    if(!m.rawmode) {
      if(i==msg.headers().end()||
	 !strcasecmp(i->second.c_str(), "text/plain")) {
	if(m.lines.size()) {
	  if(m.lines.back().attr==SLC_MAIN||
	     m.lines.back().attr==SLC_SIGNATURE||
	     m.lines.back().attr==SLC_QUOTED)
	    m.lines.push_back(tty_grin_line_t(SLC_DIVIDER, div));
	  else m.lines.push_back("");
	}
	
	unsigned int begin=0, len;
	
	bool signature=false;
	while(begin<msg.content().size()) {
	  len=0;
	  unsigned int extra=1;
	  
	  while(len<(unsigned)SLtt_Screen_Cols&&
		msg.content()[begin+len]!='\n') len++;
	  if(len>=(unsigned)SLtt_Screen_Cols) {
	    while(len>0&&!isspace(msg.content()[begin+len])) len--;
	    if(len<=0) {
	      len=SLtt_Screen_Cols;
	      extra=0;
	    }
	  }
	  
	  string line;
	  for(unsigned int i=0;i<len;i++)
	    if(msg.content()[begin+i]!='\r') line+=msg.content()[begin+i];
	  
	  if(line[0]=='-'&&line[1]=='-'&&
	     (line.size()==2||isspace(line[2]))) signature=true;
	  
	  if(signature)
	    m.lines.push_back(tty_grin_line_t(SLC_SIGNATURE, line));
	  else if(strchr(">:|", line[0]))
	    m.lines.push_back(tty_grin_line_t(SLC_QUOTED, line));
	  else m.lines.push_back(line);
	  
	  begin+=(len+extra);
	}
      } else if(i!=msg.headers().end()) {
	const char *filter=config_mime_filter(i->second.c_str());
	
	if(filter) {
	  tempfile tin, tout;
	  ofstream out(tout.filename());
	  
	  if(out.good()) {
	    for(message_part::content_t::const_iterator
		  i=msg.content().begin();
		i!=msg.content().end();++i) out<<*i;
	    out.close();
	    
	    char buf[1024];
	    snprintf(buf, sizeof(buf), "COLUMNS=%d", SLtt_Screen_Cols);
	    putenv(buf);
	    snprintf(buf, sizeof(buf), "LINES=%d", SLtt_Screen_Rows);
	    putenv(buf);
	    snprintf(buf, sizeof(buf), "%s < %s > %s",
		     filter, tout.filename(), tin.filename());
	    system(buf);
	    
	    ifstream in(tin.filename());
	    if(in.good()) {
	      if(m.lines.size()) {
		if(m.lines.back().attr==SLC_MAIN||
		   m.lines.back().attr==SLC_SIGNATURE||
		   m.lines.back().attr==SLC_QUOTED)
		  m.lines.push_back(tty_grin_line_t(SLC_DIVIDER, div));
		else m.lines.push_back("");
	      }
	      
	      in.getline(buf, sizeof(buf));
	      while(!(in.eof()||in.fail())) {
		m.lines.push_back(buf);
		in.getline(buf, sizeof(buf));
	      }
	      in.close();
	    } else report_error(strerror(errno));
	  } else report_error(strerror(errno));
	}
	m.attachments++;
      } else m.attachments++;
    }
    
    m.parts.push_back(&msg);
  }

  for(message_part::child_list::const_iterator c=msg.children().begin();
      c!=msg.children().end();++c)
    tty_grin_message_assemble(*c, m);

  if(m.rawmode&&(&msg)==&(m.msg->root()))
    for(message::raw_t::const_iterator l=m.msg->raw().begin();
	l!=m.msg->raw().end();++l)
      m.lines.push_back(*l);
}

void
tty_grin_message_uudecode(const message_part& msg,
			  tty_grin_message_t& m) {
  message_part::header_list::const_iterator
    h=msg.headers().find("content-type");

  if(h==msg.headers().end()||!strcasecmp(h->second.c_str(), "text/plain")) {
    string line;
    message_part current;
    bool decoding=false;

    for(message_part::content_t::const_iterator i=msg.content().begin();
	i!=msg.content().end();++i) {
      if(*i=='\n') {
	if(!strcasecmp(line.c_str(), "end")) {
	  m.uuencode.push_back(current);
	  decoding=false;
	}

	if(decoding) {
	  if(line[0]<33||line[0]>96) decoding=false;
	  int totalsize=(line[0]==96?0:line[0]-32);
	  for(int i=1;i+3<line.size();i+=4) {
	    char x[4];
	    for(int j=0;j<4;j++) {
	      if(line[i+j]<32||line[i+j]>96) decoding=false;
	      x[j]=line[i+j]==96?0:line[i+j]-32;
	    }
	    if(totalsize--) {
	      current.content().push_back((x[0]<<2)|((x[1]&0x30)>>4));
	      if(totalsize--) {
		current.content().push_back(((x[1]&0xf)<<4)|((x[2]&0x3c)>>2));
		if(totalsize--)
		  current.content().push_back(((x[2]&0x3)<<6)|x[3]);
	      }
	    }
	  }
	}

	if(!strncasecmp(line.c_str(), "begin ", 6)) {
	  decoding=true;
	  current.content().clear();
	  current.headers().clear();

	  const char *name=line.c_str()+6;
	  while(*name&&!isspace(*name)) name++;
	  if(*name) {
	    name++;
	    current.headers()["Content-Type"]=config_name_to_mime(name);
	    current.headers()["Content-Type.name"]=name;
	  } else current.headers()["Content-Type"]=
		   config_get("mime-type-default");
	}

	line="";
      } else if(*i!='\r'&&(line.size()||!isspace(*i))) line+=*i;
    }
    if(!strcasecmp(line.c_str(), "end")) m.uuencode.push_back(current);
  }

  for(message_part::child_list::const_iterator i=msg.children().begin();
      i!=msg.children().end();++i)
    tty_grin_message_uudecode(*i, m);
}

void
tty_grin_message_draw(const tty_grin_message_t& m) {
  list<tty_grin_line_t>::const_iterator line=m.top;
  for(int i=0;i<SLtt_Screen_Rows-2;i++) {
    SLsmg_gotorc(i, 0);
    if(line!=m.lines.end()) {
      SLsmg_set_color(line->attr);
      SLsmg_write_string((char *)line->text.c_str());
      line++;
    }
    SLsmg_erase_eol();
  }
  
  SLsmg_set_color(SLC_STATUS);
  SLsmg_fill_region(SLtt_Screen_Rows-2, 0, 1, SLtt_Screen_Cols, '-');
  SLsmg_gotorc(SLtt_Screen_Rows-2, 4);
  SLsmg_printf(_(" line %d of %d "), m.linenum+1, m.lines.size());

  if(!m.linenum) SLsmg_write_string(_("(TOP) "));
  else if(m.linenum+1>=m.lines.size()) SLsmg_write_string(_("(END) "));
  else SLsmg_printf("(%d%%) ", m.linenum*100/m.lines.size());

  if(m.attachments) SLsmg_printf(_("-- %d attachment(s) "), m.attachments);

  SLsmg_set_color(SLC_MAIN);
  SLsmg_gotorc(SLtt_Screen_Rows-1, 0);
  SLsmg_write_char(':');
  SLsmg_erase_eol();
}

void
tty_grin_message_sigwinch(int sig) {
  tty_grin_message_t& m=*(tty_grin_message_t *)sigdata_winch;

  SLtt_get_screen_size();
  SLsmg_reinit_smg();

  m.lines.clear();
  m.parts.clear();
  m.attachments=0;
  tty_grin_message_assemble(m.msg->root(), m);
  for(list<message_part>::const_iterator p=m.uuencode.begin();
      p!=m.uuencode.end();++p)
    tty_grin_message_assemble(*p, m);
  m.top=m.lines.begin();
  m.linenum=0;

  tty_grin_message_draw(m);
  SLsmg_refresh();
}

void
tty_grin_message_clearsign_check(tty_grin_message_t& m,
				 const message_part& part) {
  // This function produces weird results if the part in question isn't the
  // root part of the message.  I think it's checking the signature twice for
  // some reason.

  const char *command=config_get("pgp-clearsign-verify");

  if(!command) return;

  message_part::header_list::const_iterator
    ct=part.headers().find("content-type");
  if(ct==part.headers().end()||
     !strcasecmp(ct->second.c_str(), "text/plain")) {
    const char *lookfor="-----BEGIN PGP SIGNED MESSAGE-----",
      *look=lookfor;
    
    for(message_part::content_t::const_iterator i=part.content().begin();
	i!=part.content().end()&&*look;++i) {
      if(*i==*look) look++;
      else look=lookfor;
    }
    
    if(!*look) {
      tempfile tin, tout;
      
      ofstream out(tout.filename());
      if(out.good()) {
	for(message_part::content_t::const_iterator i=part.content().begin();
	    i!=part.content().end();++i) out<<*i;
	out.close();
	
	char buf[1024];
	const char *argv[]={tout.filename(), tin.filename()};
	pgp_printf(buf, sizeof(buf), command, 2, argv);
	system(buf);
	
	ifstream in(tin.filename());
	if(in.good()) {
	  in.getline(buf, sizeof(buf));
	  while(!(in.eof()||in.fail())) {
	    m.pgp_output.push_back(buf);
	    in.getline(buf, sizeof(buf));
	  }
	  in.close();
	} else report_error(strerror(errno));
      } else report_error(strerror(errno));
    }
  }

  for(message_part::child_list::const_iterator i=part.children().begin();
      i!=part.children().end();++i)
    tty_grin_message_clearsign_check(m, *i);
}

void
tty_grin_message_crypt_check(tty_grin_message_t& m, message_part& part) {
  const char *command=config_get("pgp-basic-decrypt");

  if(!command) return;

  message_part::header_list::const_iterator
    ct=part.headers().find("content-type");
  if(ct==part.headers().end()||
     !strcasecmp(ct->second.c_str(), "text/plain")) {
    const char *lookfor="-----BEGIN PGP MESSAGE-----",
      *look=lookfor;
    
    for(message_part::content_t::const_iterator i=part.content().begin();
	i!=part.content().end()&&*look;++i) {
      if(*i==*look) look++;
      else look=lookfor;
    }
    
    char passphrase[32];
    
    if(!*look&&read_line(-1, 0, -1, _("PGP passphrase: "),
			 passphrase, sizeof(passphrase), PASSWORD)) {
      tempfile tin, tout, tmsg;
      
      ofstream out(tout.filename());
      if(out.good()) {
	for(message_part::content_t::const_iterator i=part.content().begin();
	    i!=part.content().end();++i) out<<*i;
	out.close();
	
	int p[2];
	
	if(!pipe(p)) {
	  char buf[1024];
	  const char *argv[]={tout.filename(), tin.filename(),
			      tmsg.filename()};
	  pgp_printf(buf, sizeof(buf), command, 3, argv);
	  
	  int pid=fork();
	  
	  switch(pid) {
	  case -1:
	    close(p[0]);
	    close(p[1]);
	    report_error(strerror(errno));
	    break;
	  case 0:
	    dup2(p[0], 0);
	    close(p[1]);
	    execl("/bin/sh", "sh", "-c", buf, 0);
	    exit(-1);
	  default:
	    close(p[0]);
	    write(p[1], passphrase, strlen(passphrase));
	    close(p[1]);
	    waitpid(pid, 0, 0);
	    
	    ifstream pgp(tmsg.filename());
	    if(pgp.good()) {
	      pgp.getline(buf, sizeof(buf));
	      while(!(pgp.eof()||pgp.fail())) {
		m.pgp_output.push_back(buf);		
		pgp.getline(buf, sizeof(buf));
	      }
	      pgp.close();
	    } else report_error(strerror(errno));
	    
	    ifstream in(tin.filename());
	    if(in.good()) {
	      int c=in.get();
	      if(c>=0) {
		part.content().clear();
		part.content().push_back(c);
		while((c=in.get())>=0) part.content().push_back(c);
	      }
	      in.close();
	    } else report_error(strerror(errno));
	    
	    break;
	  }
	} else report_error(strerror(errno));
      } else report_error(strerror(errno));
    }
    
    // Just to be paranoid.
    memset(passphrase, 0, sizeof(passphrase));
  }

  for(message_part::child_list::const_iterator i=part.children().begin();
      i!=part.children().end();++i)
    tty_grin_message_clearsign_check(m, *i);
}

void
tty_grin_message_smime_sign_check(tty_grin_message_t& m) {
  // All this has to be done using the original raw message content because
  // S/MIME is so pedantic.  It'll probably break if you cough too hard.

  const char *command=config_get("pgp-verify");

  if(!command) return;

  message_part::header_list::const_iterator
    t=m.msg->root().headers().find("content-type");

  if(t!=m.msg->root().headers().end()&&
     !strcasecmp(t->second.c_str(), "multipart/signed")) {
    t=m.msg->root().headers().find("content-type.boundary");
    if(t!=m.msg->root().headers().end()) {
      tempfile content, signature, pgp;
      message::raw_t::const_iterator r=m.msg->raw().begin();

      while(r!=m.msg->raw().end()&&
	    !((*r)[0]=='-'&&(*r)[1]=='-'&&
	      !strcmp(r->c_str()+2, t->second.c_str()))) ++r;
      ++r;
      if(r!=m.msg->raw().end()) {
	ofstream out(content.filename());
	if(out.good()) {
	  message::raw_t::const_iterator re=r;
	  while(re!=m.msg->raw().end()&&
		!((*re)[0]=='-'&&(*re)[1]=='-'&&
		  !strncmp(re->c_str()+2, t->second.c_str(),
			   t->second.size()))) ++re;
	  --re;
	  while(r!=m.msg->raw().end()&&r!=re) {
	    out<<*r<<endl;
	    ++r;
	  }
	  out.close();

	  ++r;
	  while(r!=m.msg->raw().end()&&r->size()) ++r;
	  ++r;

	  ofstream sout(signature.filename());
	  if(sout.good()) {
	    while(r!=m.msg->raw().end()&&
		  !((*r)[0]=='-'&&(*r)[1]=='-'&&
		    !strncmp(r->c_str()+2, t->second.c_str(),
			     t->second.size()))) {
	      sout<<*r<<endl;
	      ++r;
	    }
	    sout.close();

	    char buf[1024];
	    const char *argv[]={content.filename(), signature.filename(),
				pgp.filename()};
	    pgp_printf(buf, sizeof(buf), command, 3, argv);
	    system(buf);

	    ifstream in(pgp.filename());
	    if(in.good()) {
	      in.getline(buf, sizeof(buf));
	      while(!(in.eof()||in.fail())) {
		m.pgp_output.push_back(buf);
		in.getline(buf, sizeof(buf));
	      }
	      in.close();
	    }
	  } else report_error(strerror(errno));
	} else report_error(strerror(errno));
      }
    }
  }
}

void
tty_grin_message_smime_crypt_check(tty_grin_message_t& m) {
  const char *command=config_get("pgp-decrypt");

  if(!command) return;
  
  message_part::header_list::const_iterator
    t=m.msg->root().headers().find("content-type");

  if(t!=m.msg->root().headers().end()&&
     !strcasecmp(t->second.c_str(), "multipart/encrypted")) {
    message_part::child_list::iterator
      e=m.msg->root().children().begin();

    while(e!=m.msg->root().children().end()) {
      t=e->headers().find("content-disposition");
      if(t==e->headers().end()||!strcasecmp(t->second.c_str(), "inline"))
	break;
      ++e;
    }

    char passphrase[32];

    if(e!=m.msg->root().children().end()&&
       read_line(-1, 0, -1, _("PGP passphrase: "),
		 passphrase, sizeof(passphrase), PASSWORD)) {
      tempfile tin, tout, tmsg;
      ofstream out(tout.filename());

      if(out.good()) {
	e->send(out, "\n");
	out.close();

	int p[2];

	if(!pipe(p)) {
	  char buf[1024];
	  const char *argv[]={tout.filename(), tin.filename(),
			      tmsg.filename()};
	  pgp_printf(buf, sizeof(buf), command, 3, argv);

	  int pid=fork();

	  switch(pid) {
	  case -1:
	    close(p[0]);
	    close(p[1]);
	    report_error(strerror(errno));
	    break;
	  case 0:
	    dup2(p[0], 0);
	    close(p[1]);
	    execl("/bin/sh", "sh", "-c", buf, 0);
	    exit(-1);
	  default:
	    close(p[0]);
	    write(p[1], passphrase, strlen(passphrase));
	    close(p[1]);
	    waitpid(pid, 0, 0);

	    ifstream pgp(tmsg.filename());
	    if(pgp.good()) {
	      pgp.getline(buf, sizeof(buf));
	      while(!(pgp.eof()||pgp.fail())) {
		m.pgp_output.push_back(buf);		
		pgp.getline(buf, sizeof(buf));
	      }
	      pgp.close();
	    } else report_error(strerror(errno));

	    ifstream in(tin.filename());
	    if(in.good()) {
	      *e=message_part(in);
	      in.close();
	    } else report_error(strerror(errno));

	    break;
	  }
	} else report_error(strerror(errno));
      } else report_error(strerror(errno));
    }
  }
}

void
tty_grin_message_pgpify(tty_grin_message_t& m) {
  tty_grin_message_smime_sign_check(m);
  tty_grin_message_smime_crypt_check(m);
  tty_grin_message_clearsign_check(m, m.msg->root());
  tty_grin_message_crypt_check(m, m.msg->root());
}

bool
tty_grin_message_command_up(tty_grin_message_t& m, const char *cmd) {
  if(m.linenum) {
    --m.top;
    --m.linenum;
  } else SLtt_beep();
  return true;
}

bool
tty_grin_message_command_down(tty_grin_message_t& m, const char *cmd) {
  if(m.linenum+1<m.lines.size()) {
    ++m.top;
    ++m.linenum;
  } else SLtt_beep();
  return true;
}

bool
tty_grin_message_command_page_down(tty_grin_message_t& m, const char *cmd) {
  if(m.linenum+1<m.lines.size()) {
    for(int i=2;i<SLtt_Screen_Rows&&m.linenum+1<m.lines.size();i++) {
      ++m.top;
      ++m.linenum;
    }
  } else SLtt_beep();
  return true;
}

bool
tty_grin_message_command_page_up(tty_grin_message_t& m, const char *cmd) {
  if(m.linenum) {
    for(int i=2;i<SLtt_Screen_Rows&&m.linenum;i++) {
      --m.top;
      --m.linenum;
    }
  } else SLtt_beep();
  return true;
}

bool
tty_grin_message_command_jump_top(tty_grin_message_t& m, const char *cmd) {
  m.top=m.lines.begin();
  m.linenum=0;
  return true;
}

bool
tty_grin_message_command_jump_end(tty_grin_message_t& m, const char *cmd) {
  --(m.top=m.lines.end());
  m.linenum=m.lines.size()-1;
  return true;
}

bool
tty_grin_message_command_prev_part(tty_grin_message_t& m, const char *cmd) {
  if(!m.rawmode&&m.linenum>1) {
    --m.top;
    --m.top;
    m.linenum-=2;
    while(m.linenum&&m.top->attr!=SLC_DIVIDER) {
      --m.top;
      --m.linenum;
    }
    if(m.linenum+1<m.lines.size()&&m.linenum) {
      ++m.top;
      ++m.linenum;
    }
  } else SLtt_beep();
  return true;
}

bool
tty_grin_message_command_next_part(tty_grin_message_t& m, const char *cmd) {
  if(!m.rawmode&&m.linenum+1<m.lines.size()) {
     ++m.top;
     ++m.linenum;
     while(m.linenum+1<m.lines.size()&&m.top->attr!=SLC_DIVIDER) {
       ++m.top;
       ++m.linenum;
     }
     if(m.linenum+1<m.lines.size()) {
       ++m.top;
       ++m.linenum;
     }
  } else SLtt_beep();
  return true;
}

bool
tty_grin_message_command_go_to_line(tty_grin_message_t& m, const char *cmd) {
  char buf[16];
  
  if(read_line(-1, 0, -1, _("Line: "), buf, sizeof(buf))) {
    unsigned int l=atoi(buf);
    if(l>0&&l+1<m.lines.size()) {
      m.linenum=l-1;
      m.top=m.lines.begin();
      while(--l) ++m.top;
    } else SLtt_beep();
  }
  return true;
}

bool
tty_grin_message_command_find(tty_grin_message_t& m, const char *cmd) {
  static char last[1024]={0};
  char next[1024];

  if(read_line(-1, 0, -1, _("Find: "), next, sizeof(next))) {
    regex_t preg;
    
    if(regcomp(&preg, next[0]?next:last,
	       REG_EXTENDED|REG_ICASE|REG_NOSUB)) SLtt_beep();
    else {
      list<tty_grin_line_t>::const_iterator t=m.top;
      unsigned int l=m.linenum;
      
      if(tolower(cmd[5])=='f') {
	++l;
	++t;
	while(l<m.lines.size()&&
	      regexec(&preg, t->text.c_str(), 0, 0, 0)) {
	  ++l;
	  ++t;
	}
      } else {
	--l;
	--t;
	while(l>0&&regexec(&preg, t->text.c_str(), 0, 0, 0)) {
	  --l;
	  --t;
	}
      }
      
      if(l>0&&l<m.lines.size()) {
	m.linenum=l;
	m.top=t;
      } else SLtt_beep();
    }
    
    if(next[0]) strcpy(last, next);
  }
  return true;
}

bool
tty_grin_message_command_attachments(tty_grin_message_t& m, const char *cmd) {
  tty_grin_message_attachments(m);
  return true;
}

bool
tty_grin_message_command_redraw(tty_grin_message_t& m, const char *cmd) {
  SLsmg_touch_lines(0, SLtt_Screen_Rows);
  return true;
}

bool
tty_grin_message_command_forward(tty_grin_message_t& m, const char *cmd) {
  message msg;
  message_part::header_list::const_iterator h;
  
  message_part body;
  body.headers()["Content-Type"]="text/plain";
  body.headers()["Content-Disposition"]="inline";
  
  prepare_message(msg.root(), body);
  msg.root().children().push_back(body);
  
  h=m.msg->root().headers().find("subject");
  if(h!=m.msg->root().headers().end())
    msg.root().headers()["Subject"]=h->second+_(" (fwd)");
  
  msg.root().headers()["Content-Type"]="multipart/mixed";
  
  char boundary[64];
  
  snprintf(boundary, sizeof(boundary), "TTY-Grin_%lx.%lx",
	   time(NULL), random());
  msg.root().headers()["Content-Type.boundary"]=boundary;
  
  message_part forward;
  forward.headers()["Content-Type"]="message/rfc822";
  forward.headers()["Content-Disposition"]="inline";
  forward.children().push_back(m.msg->root());
  msg.root().children().push_back(forward);
  
  tty_grin_compose(msg);
  return true;
}

bool
tty_grin_message_command_toggle_raw(tty_grin_message_t& m, const char *cmd) {
  m.lines.clear();
  m.parts.clear();
  m.attachments=0;
  m.rawmode=!m.rawmode;
  tty_grin_message_assemble(m.msg->root(), m);
  for(list<message_part>::const_iterator p=m.uuencode.begin();
      p!=m.uuencode.end();++p)
    tty_grin_message_assemble(*p, m);
  m.top=m.lines.begin();
  m.linenum=0;
  return true;
}

bool
tty_grin_message_command_reply(tty_grin_message_t& m, const char *cmd) {
  message msg;
  message_part::header_list::const_iterator h;

  h=m.msg->root().headers().find("message-id");
  if(h!=m.msg->root().headers().end()) {
    string ref=h->second;
    h=m.msg->root().headers().find("references");
    if(h!=m.msg->root().headers().end())
      ref=h->second+' '+ref;
    msg.root().headers()["References"]=ref;
  }
  
  h=m.msg->root().headers().find("reply-to");
  if(h==m.msg->root().headers().end())
    h=m.msg->root().headers().find("from");
  if(h!=m.msg->root().headers().end()) {
    message_part::header_list::const_iterator
      d=m.msg->root().headers().find("date");
    if(d!=m.msg->root().headers().end()) {
      char buf[1024];
      snprintf(buf, sizeof(buf), _("On %s, %s wrote:\n"),
	       d->second.c_str(), h->second.c_str());
      for(int i=0;buf[i];i++)
	msg.root().content().push_back(buf[i]);
    }
    msg.root().headers()["To"]=h->second;
  }
  
  list<const message_part *>::const_iterator
    rb=m.parts.begin();
  while(rb!=m.parts.end()) {
    message_part::header_list::const_iterator
      t=(*rb)->headers().find("content-type");
    if(t==(*rb)->headers().end()||
       !strcasecmp(t->second.c_str(), "text/plain")) break;
    ++rb;
  }
  
  if(rb!=m.parts.end()) {
    char lastchar='\n';
    for(unsigned int i=0;i<(*rb)->content().size();i++) {
      if(lastchar=='\n') msg.root().content().push_back('>');
      lastchar=(*rb)->content()[i];
      msg.root().content().push_back(lastchar);
    }
  }
  prepare_message(msg.root(), msg.root());
  
  h=m.msg->root().headers().find("subject");
  if(h!=m.msg->root().headers().end()) {
    string re;
    if(strncasecmp(h->second.c_str(), "re:", 3)) re=_("Re: ");
    re+=h->second;
    msg.root().headers()["Subject"]=re;
  }
  
  h=m.msg->root().headers().find("cc");
  if(h!=m.msg->root().headers().end()&&
     yes_or_no(_("Reply to all recipients (yes or no)? ")))
    msg.root().headers()["Cc"]=h->second;
  
  h=m.msg->root().headers().find("newsgroups");
  if(h!=m.msg->root().headers().end()&&
     yes_or_no(_("Reply to news (yes or no)? "))) {
    string ng=h->second;
    
    const newsgroup_message
      *nm=dynamic_cast<const newsgroup_message *>(m.desc);
    if(nm) {
      ng+='@';
      ng+=nm->get_newsgroup()->get_server_name();
    }
    
    msg.root().headers()["Newsgroups"]=ng;
  }

  if(m.desc) m.desc->set_flag(message_descriptor::ANSWERED);
  tty_grin_compose(msg);
  return true;
}

bool
tty_grin_message_command_save(tty_grin_message_t& m, const char *cmd) {
  mailbox *b=tty_grin_pickbox();
  if(b) {
    if(!(b->append(*m.msg, m.desc->get_flags()|
		   message_descriptor::READ)))
      report_error(_("Unable to save message"));
    delete b;
  } else SLtt_beep();
  return true;
}

bool
tty_grin_message_command_move(tty_grin_message_t& m, const char *cmd) {
  mailbox *b=tty_grin_pickbox();
  if(b) {
    if(!(b->append(*m.msg, m.desc->get_flags()|
		   message_descriptor::READ)))
      report_error(_("Unable to save message"));
    else {
      m.desc->set_flag(message_descriptor::DELETED);
      delete b;
      return false;
    }
    delete b;
  } else SLtt_beep();
  return true;
}

bool
tty_grin_message_command_export(tty_grin_message_t& m, const char *cmd) {
  char buf[1024];

  if(read_line(-1, 0, -1, _("Filename: "), buf, sizeof(buf), FILENAME)&&
     (access(buf, F_OK)||
      yes_or_no(_("Overwrite existing file (yes or no)? ")))) {
    ofstream out(buf);
    
    if(out.good()) {
      m.msg->send_original(out, "\n");
      out.close();
    } else report_error(strerror(errno));
  }
  return true;
}

bool
tty_grin_message_command_delete(tty_grin_message_t& m, const char *cmd) {
  m.desc->set_flag(message_descriptor::DELETED);
  return false;
}

bool
tty_grin_message_command_quit(tty_grin_message_t& m, const char *cmd) {
  return false;
}

bool
tty_grin_message_command_help(tty_grin_message_t& m, const char *cmd) {
  static tty_grin_help_key_t help_keys[]={
    {"message", "prev-part",    N_("Previous part")},
    {"message", "next-part",    N_("Next part")},
    {"message", "go-to-line",   N_("Goto line")},
    {"default", "find-forward", N_("Search forward")},
    {"default", "find-back",    N_("Search back")},
    {"message", "attachments",  N_("Attachments")},
    {"message", "forward",      N_("Forward")},
    {"message", "reply",        N_("Reply")},
    {"message", "save",         N_("Save (to mailbox)")},
    {"message", "export",       N_("Save (to file)")},
    {"message", "move",         N_("Move")},
    {"message", "delete",       N_("Delete")},
    {"message", "toggle-raw",   N_("Raw mode")},
    {"default", "quit",         N_("Back")},
    {0, 0, 0}
  };

  static tty_grin_help_t help={
    N_("Message viewer"),
    N_("Read messages that have been sent to you or to subscribed newsgroups. "
       "Replying or forwarding a message will start a message composition."),
    help_keys
  };

  tty_grin_help(help);
  return true;
}

void
tty_grin_message(message msg, message_descriptor *desc) {
  tty_grin_message_t m;

  m.parts.clear();
  m.attachments=0;
  m.msg=&msg;
  m.desc=desc;
  m.rawmode=false;

  tty_grin_message_pgpify(m);
  tty_grin_message_uudecode(msg.root(), m);
  tty_grin_message_assemble(msg.root(), m);
  for(list<message_part>::const_iterator p=m.uuencode.begin();
      p!=m.uuencode.end();++p)
    tty_grin_message_assemble(*p, m);

  m.top=m.lines.begin();
  m.linenum=0;

  void *sigdata_old=sigdata_winch;
  SLSig_Fun_Type *f=SLsignal(SIGWINCH, tty_grin_message_sigwinch);
  sigdata_winch=&m;

  for(;;) {
    tty_grin_message_draw(m);
    SLsmg_refresh();

    const char *cmd=kbd_translate("message", SLkp_getkey());

    static struct {
      const char *command;
      bool (*func)(tty_grin_message_t&, const char *);
    } commands[]={
      {"up",           tty_grin_message_command_up},
      {"down",         tty_grin_message_command_down},
      {"page-up",      tty_grin_message_command_page_up},
      {"page-down",    tty_grin_message_command_page_down},
      {"jump-top",     tty_grin_message_command_jump_top},
      {"jump-end",     tty_grin_message_command_jump_end},
      {"next-part",    tty_grin_message_command_next_part},
      {"prev-part",    tty_grin_message_command_prev_part},
      {"go-to-line",   tty_grin_message_command_go_to_line},
      {"find-forward", tty_grin_message_command_find},
      {"find-back",    tty_grin_message_command_find},
      {"attachments",  tty_grin_message_command_attachments},
      {"redraw",       tty_grin_message_command_redraw},
      {"forward",      tty_grin_message_command_forward},
      {"toggle-raw",   tty_grin_message_command_toggle_raw},
      {"reply",        tty_grin_message_command_reply},
      {"save",         tty_grin_message_command_save},
      {"move",         tty_grin_message_command_move},
      {"export",       tty_grin_message_command_export},
      {"help",         tty_grin_message_command_help},
      {"delete",       tty_grin_message_command_delete},
      {"quit",         tty_grin_message_command_quit},
      {0, 0}
    };

    if(cmd) {
      int i=0;

      while(commands[i].command&&strcasecmp(commands[i].command, cmd)) i++;

      if(commands[i].command) {
	if(!commands[i].func(m, cmd)) {
	  sigdata_winch=sigdata_old;
	  SLsignal(SIGWINCH, f);
	  return;
	}
      } else SLtt_beep();
    } else SLtt_beep();
  }
}
