// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <fstream>
#include <errno.h>
#include "localboxlist.h"
#include "imap4boxlist.h"
#include "localmailbox.h"
#include "maildir.h"
#include "grin.h"
#include "ctree.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace std;

static tty_grin_help_key_t
tty_grin_pickbox_help_keys[]={
  {"tree",    "collapse",     N_("Collapse")},
  {"tree",    "expand",       N_("Expand")},
  {"tree",    "collapse-all", N_("Collapse all")},
  {"tree",    "expand-all",   N_("Expand all")},
  {"default", "quit",         N_("Back")},
  {0, 0, 0}
};

static tty_grin_help_t tty_grin_pickbox_help={
  N_("Select mailbox"),
  N_("Select a mailbox into which a message will be saved."),
  tty_grin_pickbox_help_keys
};

class tty_grin_pickbox_boxlist_t : public ctree {
public:
  tty_grin_pickbox_boxlist_t(boxlist&);

  mailbox *get_result(void) { return result; }

protected:
  void draw(int, int, int, int);
  void winch(void);
  bool keypress(int);

private:
  boxlist& bl;
  mailbox *result;
};

tty_grin_pickbox_boxlist_t::tty_grin_pickbox_boxlist_t(boxlist& b) :
  ctree(1), bl(b) {
  for(int i=0;i<bl.size();i++) 
    (new ctree_node(*this))->set_text(0, bl[i]);

  set_width(0, SLtt_Screen_Cols);
  result=0;
}

void
tty_grin_pickbox_boxlist_t::draw(int r, int c, int rows, int columns) {
  SLsmg_set_color(SLC_STATUS);
  SLsmg_fill_region(r+rows-2, c, 1, columns, '-');
  SLsmg_gotorc(r+rows-2, c+4);
  SLsmg_write_string(_(" Save to mailbox "));

  SLsmg_set_color(SLC_MAIN);
  SLsmg_fill_region(r+rows-1, c, 1, columns, ' ');

  ctree::draw(r, c, rows-2, columns);
}

void
tty_grin_pickbox_boxlist_t::winch(void) {
  SLtt_get_screen_size();
  set_width(0, SLtt_Screen_Cols);
  ctree::winch();
}

bool
tty_grin_pickbox_boxlist_t::keypress(int key) {
  const char *cmd=kbd_translate("pickbox", key);

  if(cmd) {
    if(!strcasecmp(cmd, "select")) {
      if(get_cursor()) {
	result=bl.openbox(get_cursor()->get_text(0));
	if(result) {
	  localmailbox *b=dynamic_cast<localmailbox *>(result);
	  if(!b||b->lock()) return false;
	  else {
	    delete result;
	    result=0;
	    report_error(_("Mailbox is currently in use"));
	  }
	} else report_error(strerror(errno));
      }
    } else if(!strcasecmp(cmd, "help"))
      tty_grin_help(tty_grin_pickbox_help);
    else return ctree::keypress(key);
  } else return ctree::keypress(key);

  return true;
}

class tty_grin_pickbox_imap4_t : public ctree {
public:
  tty_grin_pickbox_imap4_t(void);

  mailbox *get_result(void) { return result; }

protected:
  void draw(int, int, int, int);
  void winch(void);
  bool keypress(int);

private:
  mailbox *result;
};

tty_grin_pickbox_imap4_t::tty_grin_pickbox_imap4_t(void) : ctree(4) {
  string path(getenv("HOME"));
  path+="/.ttygrin/imap4";
  ifstream in(path.c_str());

  if(in.good()) {
    char buf[1024];

    in.getline(buf, sizeof(buf));
    while(!(in.eof()||in.fail())) {
      char *x=buf;
      ctree_node *n=new ctree_node(*this);

      for(int i=0;i<4;i++) {
	char *y=strsep(&x, "\t");
	if(y) n->set_text(i, y);
      }
      in.getline(buf, sizeof(buf));
    }
    in.close();
  }

  set_width(0, SLtt_Screen_Cols/3);
  set_width(1, SLtt_Screen_Cols/3);
  set_width(2, SLtt_Screen_Cols/3);

  result=0;
}

void
tty_grin_pickbox_imap4_t::winch(void) {
  SLtt_get_screen_size();
  set_width(0, SLtt_Screen_Cols/3);
  set_width(1, SLtt_Screen_Cols/3);
  set_width(2, SLtt_Screen_Cols/3);
  ctree::winch();
}

void
tty_grin_pickbox_imap4_t::draw(int r, int c, int rows, int columns) {
  SLsmg_set_color(SLC_STATUS);
  SLsmg_fill_region(r+rows-2, c, 1, columns, '-');
  SLsmg_gotorc(r+rows-2, c+4);
  SLsmg_write_string(_(" Save to mailbox -- IMAP4 "));

  SLsmg_set_color(SLC_MAIN);
  SLsmg_fill_region(r+rows-1, c, 1, columns, ' ');

  ctree::draw(r, c, rows-2, columns);
}

bool
tty_grin_pickbox_imap4_t::keypress(int key) {
  const char *cmd=kbd_translate("pickbox", key);

  if(cmd) {
    if(!strcasecmp(cmd, "select")) {
      if(get_cursor()) {
	imap4boxlist bl(get_cursor()->get_text(0),
			get_cursor()->get_text(1),
			get_cursor()->get_text(2),
			get_cursor()->get_text(3));
	tty_grin_pickbox_boxlist_t blt(bl);
	
	blt.set_cursor_color(SLC_CURSOR);
	blt.run();
	
	if(blt.get_result()) {
	  result=blt.get_result();
	  return false;
	}
      }
    } else if(!strcasecmp(cmd, "help"))
      tty_grin_help(tty_grin_pickbox_help);
    else return ctree::keypress(key);
  } else return ctree::keypress(key);

  return true;
}

class tty_grin_pickbox_t : public ctree {
public:
  tty_grin_pickbox_t(void);

  mailbox *get_result(void) { return result; }

protected:
  void draw(int, int, int, int);
  void winch(void);
  bool keypress(int);

private:
  mailbox *result;
};

tty_grin_pickbox_t::tty_grin_pickbox_t(void) : ctree(1) {
  (new ctree_node(*this))->set_text(0, _("Inbox"));
  (new ctree_node(*this))->set_text(0, _("Local mailboxes"));
  (new ctree_node(*this))->set_text(0, _("IMAP4"));

  set_width(0, SLtt_Screen_Cols);
  result=0;
}

void
tty_grin_pickbox_t::winch(void) {
  SLtt_get_screen_size();
  set_width(0, SLtt_Screen_Cols);
  ctree::winch();
}

void
tty_grin_pickbox_t::draw(int r, int c, int rows, int columns) {
  SLsmg_set_color(SLC_STATUS);
  SLsmg_fill_region(r+rows-2, c, 1, columns, '-');
  SLsmg_gotorc(r+rows-2, c+4);
  SLsmg_write_string(_(" Save to mailbox "));

  SLsmg_set_color(SLC_MAIN);
  SLsmg_fill_region(r+rows-1, c, 1, columns, ' ');

  ctree::draw(r, c, rows-2, columns);
}

bool
tty_grin_pickbox_t::keypress(int key) {
  const char *cmd=kbd_translate("pickbox", key);

  if(cmd) {
    if(!strcasecmp(cmd, "select")) {
      if(!(get_cursor()->get_next())) {
	tty_grin_pickbox_imap4_t im;
	
	im.set_cursor_color(SLC_CURSOR);
	im.run();
	
	if(im.get_result()) {
	  result=im.get_result();
	  return false;
	}
      } else if(get_cursor()->get_prev()) {
	localboxlist b(config_get("mail"));
	tty_grin_pickbox_boxlist_t bl(b);
	
	bl.set_cursor_color(SLC_CURSOR);
	bl.run();
	
	if(bl.get_result()) {
	  result=bl.get_result();
	  return false;
	}
      } else {
	const char *md=config_get("maildir");
	
	if(md) {
	  result=new maildir(md);
	  return false;
	} else {
	  result=new localmailbox(config_get("inbox"));
	  if(((localmailbox *)result)->lock()) return false;
	  delete result;
	  result=0;
	  report_error(_("Mailbox is currently in use"));
	}
      }
    } else if(!strcasecmp(cmd, "help"))
      tty_grin_help(tty_grin_pickbox_help);
    else return ctree::keypress(key);
  } else return ctree::keypress(key);

  return true;
}

mailbox *
tty_grin_pickbox(void) {
  tty_grin_pickbox_t pb;

  pb.set_cursor_color(SLC_CURSOR);
  pb.run();

  return pb.get_result();
}
